import os.path
import epub_meta
from PyPDF2 import PdfFileReader
from typing import Optional, List

from PyQt5.QtCore import QRect

try:
    import fitz
    has_fitz = True
except ImportError:
    has_fitz = False

import logging as _logging

logger = _logging.getLogger(__name__)


class PageImage():
    def __init__(self, file_doc, page_idx: int, zoom: Optional[float] = None):
        if not has_fitz:
            return None

        if file_doc is None or page_idx > file_doc.number_of_pages() - 1:
            return None

        self.zoom = zoom or file_doc._zoom
        self.pixmap: fitz.Pixmap = file_doc.file_doc[page_idx] \
                                           .get_pixmap(
                                               matrix=fitz.Matrix(self.zoom, self.zoom),
                                               alpha=False)

        self.image_bytes = self.pixmap.tobytes("ppm")

        self.width = self.pixmap.width
        self.height = self.pixmap.height
        self.stride = self.pixmap.stride

    def set_pixmap(self, pixmap: fitz.Pixmap):
        self.pixmap = pixmap
        self.image_bytes = self.pixmap.tobytes("ppm")


class FileDoc():
    def __init__(self, filepath):

        if not os.path.exists(filepath):
            logger.error(f"File not found: {filepath}")

        self.filepath = filepath
        name, ext = os.path.splitext(self.filepath)
        self.file_name = name
        self.file_ext = ext.lower()
        self.title = ""
        self.author = ""
        self.select_annot_id = None

        if has_fitz:
            self.file_doc = fitz.open(filepath)
        elif self.file_ext == ".pdf":
            self.file_doc = PdfFileReader(open(self.filepath, "rb"))
        else:
            self.file_doc = None

        self._current_idx: int = 0
        self._zoom = None
        self._matrix = None

        self.current_page_image = None
        self.set_zoom(1.5)

        self._parse_metadata()

    def set_zoom(self, x):
        self._zoom = x
        if has_fitz:
            self._matrix = fitz.Matrix(self._zoom, self._zoom)
            self.current_page_image = self.current_unchanged_page_image()

    def number_of_pages(self) -> int:
        if has_fitz and self.file_doc:
            return len(self.file_doc)
        else:
            return 0

    def go_to_page(self, page: int):
        if page <= 0:
            return

        if self._current_idx == page - 1:
            return

        self._current_idx = page - 1

        self.current_page_image = self.current_unchanged_page_image()

    def get_selection_rect(self, select_rect: QRect) -> Optional[fitz.Rect]:
        if not self.file_doc:
            return None

        page = self.current_page()
        page_image = self.current_unchanged_page_image()
        page_rect: fitz.Rect = page.rect

        s = select_rect.getCoords()

        # increase the selection area for better results
        c = [0, 0, 0, 0]
        c[0] = s[0] - 5
        c[1] = s[1] - 10
        c[2] = s[2] + 3
        c[3] = s[3] + 10

        # transform image_rect pixel coords to page_rect coords
        w_scale = page_rect.width / page_image.width
        h_scale = page_rect.height / page_image.height

        rect = fitz.Rect(w_scale * c[0], h_scale * c[1], w_scale * c[2], h_scale * c[3])

        return rect

    def get_selection_quads(self, select_rect: QRect) -> Optional[List[fitz.Quad]]:
        if not self.file_doc:
            return None

        page = self.current_page()
        rect = self.get_selection_rect(select_rect)
        text = page.get_textbox(rect)

        # using quads option
        quads = page.search_for(text, quads=True)

        if quads:
            # keep those text matches which are inside the selection area
            quads = list(filter(lambda q: rect.intersects(q.rect), quads))
        else:
            return None

        return quads

    def paint_selection_on_current(self, select_rect: QRect):
        if not self.file_doc:
            return None

        quads = self.get_selection_quads(select_rect)
        if not quads:
            self.current_page_image = self.current_unchanged_page_image()
            return

        if self.file_ext == ".pdf":
            # PDF supports drawing shapes.
            # We create a temp page to render drawings.

            self.file_doc.fullcopy_page(self._current_idx, -1)
            temp_page_idx = len(self.file_doc) - 1
            page = self.file_doc[temp_page_idx]

            shape: fitz.Shape = page.new_shape()
            orange = (0.913, 0.329, 0.125)  # rgb(233, 84, 32)

            for q in quads:
                rect = q.rect

                shape.draw_rect(rect)
                shape.finish(fill=orange)

            shape.commit(overlay=False)

            self.current_page_image = self.page_image(temp_page_idx)
            self.file_doc.delete_page(temp_page_idx)

        else:
            page = self.current_page()

            page_pixmap: fitz.Pixmap = page.get_pixmap(matrix=self._matrix, alpha=False)

            # page_pixmap: fitz.Pixmap = page.get_pixmap(matrix=self._matrix, alpha=True)
            # alphas = bytearray([255] * (page_pixmap.width * page_pixmap.height))
            # sel_pixmap: fitz.Pixmap()

            for q in quads:
                rect = q.rect
                rect.transform(self._matrix)

                page_pixmap.invert_irect(rect)

                # orange = (233, 84, 32)
                # page_pixmap.set_rect(rect, orange)

            # data = page_pixmap.tobytes("ppm")
            # pix = fitz.Pixmap()

            page_image = self.current_unchanged_page_image()
            page_image.set_pixmap(page_pixmap)
            self.current_page_image = page_image

    def get_selection_text(self, select_rect: QRect) -> Optional[str]:
        if not self.file_doc:
            return None

        page = self.current_page()
        rect = self.get_selection_rect(select_rect)
        text = page.get_textbox(rect)
        return text

    def select_highlight_text(self, select_rect: QRect) -> Optional[str]:
        if not self.file_doc:
            return None

        page = self.current_page()
        rect = self.get_selection_rect(select_rect)
        text = page.get_textbox(rect)

        quads = self.get_selection_quads(select_rect)

        if not quads:
            return text

        if self.select_annot_id:
            for a in page.annots():
                if a.info['id'] == self.select_annot_id:
                    page.delete_annot(a)

        annot = page.add_highlight_annot(quads=quads)
        if annot:
            self.select_annot_id = annot.info['id']

        return text

    def current_page_number(self) -> int:
        return self._current_idx + 1

    def current_page_idx(self) -> int:
        return self._current_idx

    def current_page(self) -> fitz.Page:
        # if not self.file_doc:
        #     raise ...
        return self.file_doc[self._current_idx]

    def current_unchanged_page_image(self) -> Optional[PageImage]:
        return self.page_image(self._current_idx)

    def page_image(self, page_idx: int, zoom: Optional[float] = None) -> Optional[PageImage]:
        if not has_fitz:
            return None

        return PageImage(self, page_idx, zoom)

    def _parse_metadata(self):
        if not self.file_doc:
            return

        if has_fitz:
            self.title = self.file_doc.metadata['title']
            self.author = self.file_doc.metadata['author']

        elif self.file_ext == ".pdf":
            info = self.file_doc.getDocumentInfo()
            self.title = info.title
            self.author = info.author

        elif self.file_ext == ".epub":
            info = epub_meta.get_epub_metadata(self.filepath, read_cover_image=False, read_toc=False)
            self.title = info['title']
            self.author = info['authors'][0]



import sys
import logging as _logging
from functools import partial
import pyperclip

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (QWidget, QVBoxLayout, QHBoxLayout, QTextBrowser,
                             QPushButton, QLabel, QMainWindow, QSizePolicy)

logger = _logging.getLogger(__name__)


class ErrorMessageWindow(QMainWindow):
    def __init__(self, user_message=None, debug_info=None, status=None) -> None:
        super().__init__()
        self.setWindowTitle("Application Error")
        self.setFixedSize(500, 500)
        self.setWindowFlags(Qt.WindowType.WindowStaysOnTopHint)

        self._central_widget = QWidget(self)
        self.setCentralWidget(self._central_widget)

        self._layout = QVBoxLayout()
        self._central_widget.setLayout(self._layout)

        self._msg = QLabel()
        self._msg.setWordWrap(True)

        if user_message:
            self._msg.setText(user_message)
        else:
            self._msg.setText("<p>The application encountered and error.</p>")

        self._msg.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self._layout.addWidget(self._msg)

        if debug_info:
            debug_info = f"```\n{debug_info}\n```"
            self._info_help = QLabel()
            self._info_help.setText("""
<p>
  You can submit an issue report at:<br>
  <a href="https://github.com/simsapa/simsapa/issues">https://github.com/simsapa/simsapa/issues</a>
</p>
<p>
  Or send an email to <a href="mailto:profound.labs@gmail.com">profound.labs@gmail.com</a>
</p>
<p>
  Please copy the error mesage in your bug report:
</p>""")

            self._info_help.setAlignment(Qt.AlignmentFlag.AlignCenter)
            self._info_help.setOpenExternalLinks(True)
            self._layout.addWidget(self._info_help)

            self._copy_button = QPushButton("Copy")
            self._copy_button.setFixedSize(100, 30)

            self._copy_button.clicked.connect(partial(self._copy, debug_info))

            layout = QHBoxLayout()
            layout.addWidget(self._copy_button)

            self._layout.addLayout(layout)

            self._debug_info = QTextBrowser()
            self._debug_info.setText(debug_info)
            self._debug_info.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
            self._layout.addWidget(self._debug_info)

        buttons_layout = QHBoxLayout()
        buttons_layout.setContentsMargins(0, 20, 0, 10)

        self._continue_button = QPushButton("Continue")
        self._continue_button.setFixedSize(100, 30)
        self._quit_button = QPushButton("Quit")
        self._quit_button.setFixedSize(100, 30)

        self._continue_button.clicked.connect(partial(self.close))
        self._quit_button.clicked.connect(partial(self._close_and_quit))

        buttons_layout.addWidget(self._continue_button)
        buttons_layout.addWidget(self._quit_button)

        self._layout.addLayout(buttons_layout)

    def _copy(self, text):
        pyperclip.copy(text)

    def _close_and_quit(self):
        self.close()
        status = 1
        logger.error(f"Exiting with status {status}.")
        sys.exit(status)



import os
from pathlib import Path
from typing import Dict
from queue import Queue
from dotenv import load_dotenv
import appdirs

load_dotenv()

SIMSAPA_PACKAGE_DIR = Path(os.path.dirname(__file__)).absolute()

ALEMBIC_INI = SIMSAPA_PACKAGE_DIR.joinpath('alembic.ini')
ALEMBIC_DIR = SIMSAPA_PACKAGE_DIR.joinpath('alembic')

SIMSAPA_DIR = Path(appdirs.user_data_dir('simsapa'))

TEST_ASSETS_DIR = SIMSAPA_PACKAGE_DIR.joinpath('../tests/data/assets')

s = os.getenv('USE_TEST_DATA')
if s is not None and s.lower() == 'true':
    ASSETS_DIR = TEST_ASSETS_DIR
else:
    ASSETS_DIR = SIMSAPA_DIR.joinpath('assets')

INDEX_DIR = ASSETS_DIR.joinpath('index')

APP_DB_PATH = ASSETS_DIR.joinpath('appdata.sqlite3')
USER_DB_PATH = ASSETS_DIR.joinpath('userdata.sqlite3')

APP_QUEUES: Dict[str, Queue] = {}



"""Stardict format related funcions

Stardict format documentation:
https://github.com/huzheng001/stardict-3/blob/master/dict/doc/StarDictFileFormat

Notes about StarDict dictionary format
http://dhyannataraj.github.io/blog/2010/10/04/Notes-about-stardict-dictionry-format/

Support functions:
https://github.com/codito/stargaze/blob/master/stargaze.py
"""

import logging as _logging
from pathlib import Path
import datetime
from typing import List, TypedDict, Optional
import shutil
from zipfile import ZipFile
import struct
import idzip

from simsapa import SIMSAPA_DIR

logger = _logging.getLogger(__name__)

class DictError(Exception):
    """Error in the dictionary."""
    pass

class StarDictIfo(TypedDict):
    """
    Available options:

    ```
    bookname=      // required
    wordcount=     // required
    synwordcount=  // required if ".syn" file exists.
    idxfilesize=   // required
    idxoffsetbits= // New in 3.0.0
    author=
    email=
    website=
    description= // You can use <br> for new line.
    date=
    sametypesequence= // very important.
    dicttype=
    ```

    sametypesequence=m The data should be a utf-8 string ending with '\\0'.
    sametypesequence=h Html


    Example contents of an .ifo file:

    ```
    StarDict's dict ifo file
    version=3.0.0
    bookname=Digital Pāli Dictionary
    wordcount=36893
    synwordcount=1727042
    idxfilesize=747969
    idxoffsetbits=32
    author=Digital Pāli Tools <digitalpalitools@gmail.com>
    website=https://github.com/digitalpalitools
    description=The next generation comprehensive Digital Pāli Dictionary.
    date=2021-10-31T08:56:25Z
    sametypesequence=h
    ```
    """

    version: str
    bookname: str
    wordcount: str
    synwordcount: str
    idxfilesize: str
    idxoffsetbits: str
    author: str
    email: str
    website: str
    description: str
    date: str
    sametypesequence: str
    dicttype: str

def ifo_from_opts(opts: dict[str, str]) -> StarDictIfo:
    ifo = StarDictIfo(
        version = '',
        bookname = '',
        wordcount = '',
        synwordcount = '',
        idxfilesize = '',
        idxoffsetbits = '',
        author = '',
        email = '',
        website = '',
        description = '',
        date = '',
        sametypesequence = '',
        dicttype = '',
    )
    for k in opts.keys():
        if k in ifo.keys():
            ifo[k] = opts[k]

    return ifo

class IdxEntry(TypedDict):
    # a utf-8 string, `\0` stripped
    word: str
    # word data's offset in .dict file
    offset_begin: int
    # word data's total size in .dict file
    data_size: int

class DictEntry(TypedDict):
    # a utf-8 string, `\0` stripped
    word: str
    definition_plain: str
    definition_html: str
    synonyms: List[str]

SynEntries = dict[
    # a utf-8 string, `\0` stripped
    str,
    # indices from the .idx
    List[int],
]

class StarDictPaths(TypedDict):
    zip_path: Path
    unzipped_dir: Path
    icon_path: Optional[Path]
    ifo_path: Optional[Path]
    idx_path: Optional[Path]
    dic_path: Optional[Path]
    syn_path: Optional[Path]

def new_stardict_paths(zip_path: Path):
    unzipped_dir: Path = SIMSAPA_DIR.joinpath("unzipped_stardict")
    return StarDictPaths(
        zip_path = zip_path,
        unzipped_dir = unzipped_dir,
        icon_path = None,
        ifo_path = None,
        idx_path = None,
        dic_path = None,
        syn_path = None,
    )

def parse_stardict_zip(zip_path: Path) -> StarDictPaths:

    stardict_paths = new_stardict_paths(zip_path)
    unzipped_dir = stardict_paths['unzipped_dir']

    # Find the .ifo, .idx, .dic, .syn
    hits = {'*.ifo': [], '*.idx': [], '*.dic*': [], '*.syn': []}
    try:
        # delete and re-create to make sure it's an empty directory
        if unzipped_dir.exists():
            shutil.rmtree(unzipped_dir)
        unzipped_dir.mkdir()

        with ZipFile(zip_path, 'r') as z:
            z.extractall(unzipped_dir)

        # NOTE: The zip file may or may not have a top-level folder. A
        # dictionary may be compressed as '*.dict.dz'.

        for ext in hits.keys():
            a = list(unzipped_dir.glob(f"**/{ext}"))
            if len(a) == 0:
                # .syn is optional
                if ext == '*.syn':
                    hits[ext] = [None]
                else:
                    msg = f"ERROR: Can't find this type of file in the .zip: {ext}"
                    logger.error(msg)
                    raise DictError(msg)
            else:
                hits[ext] = [a[0]]

    except Exception as e:
        logger.error(e)
        raise e

    stardict_paths['ifo_path'] = hits['*.ifo'][0]
    stardict_paths['idx_path'] = hits['*.idx'][0]
    stardict_paths['dic_path'] = hits['*.dic*'][0]
    stardict_paths['syn_path'] = hits['*.syn'][0]

    return stardict_paths

def parse_ifo(paths: StarDictPaths) -> StarDictIfo:
    if paths['ifo_path'] is None:
        msg = f"ifo file is None"
        logger.error(msg)
        raise DictError(msg)

    ifo_path = paths['ifo_path']

    if not ifo_path.exists():
        msg = f"ifo file not found: {ifo_path}"
        logger.error(msg)
        raise DictError(msg)

    magic_header = "StarDict's dict ifo file"
    with open(ifo_path, 'r', encoding='utf-8') as f:
        magic_string = f.readline().rstrip()
        if (magic_string != magic_header):
            msg = "IFO: Incorrect header: {}".format(magic_string)
            logger.error(msg)
            raise DictError(msg)

        a = (l.split("=") for l in map(str.rstrip, f) if l != "")
        opts = {k.strip(): v.strip() for k, v in a}

        return ifo_from_opts(opts)

def stardict_to_dict_entries(paths: StarDictPaths) -> List[DictEntry]:

    idx = parse_idx(paths)
    ifo = parse_ifo(paths)
    syn = parse_syn(paths)
    words = parse_dict(paths, ifo, idx, syn)

    return words

def parse_idx(paths: StarDictPaths) -> List[IdxEntry]:
    """Parse an .idx file."""

    if paths['idx_path'] is None:
        msg = f"idx file is None"
        logger.error(msg)
        raise DictError(msg)

    idx_path = paths['idx_path']

    if not idx_path.exists():
        msg = f"idx file not found: {idx_path}"
        logger.error(msg)
        raise DictError(msg)

    words_index = []

    # idxoffsetbits can be 32bit or 64bit.
    # offset refers to two integers.
    # struct packing uses ">II", two unsigned ints, 2*4 bytes = 32 bits
    offset_size = 8
    with open(idx_path, "rb") as f:
        while True:
            word_str = _read_word(f).rstrip('\0')

            word_pointer = f.read(offset_size)
            if not word_pointer:
                break

            offset_begin, data_size = struct.unpack(">II", word_pointer)
            words_index.append(IdxEntry(
                word = word_str,
                offset_begin = offset_begin,
                data_size = data_size,
            ))

    return words_index

def parse_dict(paths: StarDictPaths,
               ifo: StarDictIfo,
               idx: List[IdxEntry],
               syn: Optional[SynEntries]) -> List[DictEntry]:
    """Parse a .dict file."""

    dict_path = paths['dic_path']

    if dict_path is None:
        msg = f"dict file is None"
        logger.error(msg)
        raise DictError(msg)

    if not dict_path.exists():
        msg = f"dict file not found: {dict_path}"
        logger.error(msg)
        raise DictError(msg)

    words: List[DictEntry] = []

    open_dict = open
    if f"{dict_path}".endswith(".dz"):
        open_dict = idzip.open

    if len(ifo["sametypesequence"]) > 0:
        types = ifo["sametypesequence"]
    else:
        types = "m"

    with open_dict(dict_path, "rb") as f:
        for i in idx:

            dict_word = i['word']
            f.seek(i["offset_begin"])
            data = f.read(i["data_size"])
            data_str: str = data.decode("utf-8").rstrip("\0")

            definition_plain = ""
            definition_html = ""
            # Only accept sametypesequence = m, h
            if types == "m" and data_str[0] == "m":
                definition_plain = data_str[1:]

            if types == "h":
                definition_html = data_str

            synonyms = []
            if syn is not None and dict_word in syn.keys():
                synonyms = list(map(lambda x: idx[x]['word'], syn[dict_word]))
                # FIXME synonyms seem to just repeat the dict word
                # if len(synonyms) > 0:
                #     print(syn[dict_word])
                #     print(synonyms)

            words.append(DictEntry(
                word = dict_word,
                definition_plain = definition_plain,
                definition_html = definition_html,
                synonyms = synonyms,
            ))

    return words

def parse_syn(paths: StarDictPaths) -> Optional[SynEntries]:
    """Parse a .syn file with synonyms.

    The .syn format:

    Each item contains one string and a number.

    synonym_word;  // a utf-8 string terminated by NUL '\\0'.
    original_word_index; // original word's index in .idx file.

    Then other items without separation.

    When you input synonym_word, StarDict will search original_word; The length
    of "synonym_word" should be less than 256. In other words, (strlen(word) <
    256).

    original_word_index is a 32-bits unsigned number in network byte order. Two
    or more items may have the same "synonym_word" with different
    original_word_index.
    """

    if paths['syn_path'] is None:
        # Syn file is optional
        return None

    syn_path = paths['syn_path']

    if not syn_path.exists():
        msg = f"syn file not found: {syn_path}"
        logger.error(msg)
        raise DictError(msg)

    syn_entries: SynEntries = {}

    with open(syn_path, 'rb') as f:
        while True:
            word_str = _read_word(f).rstrip('\0')

            word_pointer = f.read(4)
            if not word_pointer:
                break

            if word_str not in syn_entries.keys():
                syn_entries[word_str] = []

            syn_entries[word_str].extend(struct.unpack(">I", word_pointer))

    return syn_entries

def _read_word(f) -> str:
    r"""Read a unicode NUL `\\0` terminated string from a file like object."""

    word = bytearray()
    c = f.read(1)
    while c and c != b'\0':
        word.extend(c)
        c = f.read(1)
    return word.decode('utf-8')

def write_ifo(ifo: StarDictIfo, paths: StarDictPaths):
    """Writes .ifo"""

    if paths['ifo_path'] is None:
        logger.error("ifo_path is required")
        return

    lines: List[str] = ["StarDict's dict ifo file"]

    required = ['bookname', 'wordcount', 'synwordcount', 'idxfilesize', 'sametypesequence']
    missing = []
    for k in required:
        if k not in ifo.keys() and len(ifo[k]) > 0 and ifo[k] != 'None':
            missing.append(k)

    if len(missing) > 0:
        logger.error(f"Missing required keys: {missing}")
        return

    for k in ifo.keys():
        v = ifo[k]
        if len(v) > 0 and v != 'None':
            lines.append(f"{k}={v}")

    with open(paths['ifo_path'], 'w') as f:
        f.write("\n".join(lines))

class WriteResult(TypedDict):
    idx_size: Optional[int]
    syn_count: Optional[int]

def write_words(words: List[DictEntry], paths: StarDictPaths) -> WriteResult:
    """Writes .idx, .dict.dz, .syn"""

    res = WriteResult(
        idx_size = None,
        syn_count = None
    )

    if paths['idx_path'] is None or paths['dic_path'] is None:
        logger.error("idx_path and dic_path are required")
        return res

    idx: List[IdxEntry] = []

    with idzip.IdzipFile(f"{paths['dic_path']}", "wb") as f:
        offset_begin = 0
        data_size = 0
        for w in words:
            d = bytes(w['definition_html'], 'utf-8')
            f.write(d)

            data_size = len(d)

            idx.append(IdxEntry(
                word = w['word'],
                offset_begin = offset_begin,
                data_size = data_size,
            ))

            offset_begin += data_size

    with open(paths['idx_path'], 'wb') as f:
        for i in idx:
            d = bytes(f"{i['word']}\0", "utf-8")
            f.write(d)
            d = struct.pack(">II", i['offset_begin'], i['data_size'])
            f.write(d)

    res['idx_size'] = paths['idx_path'].stat().st_size

    if paths['syn_path'] is not None:
        res['syn_count'] = 0

        with open(paths['syn_path'], 'wb') as f:
            for n, w in enumerate(words):

                if res['syn_count'] is not None:
                    res['syn_count'] += len(w['synonyms'])

                for s in w['synonyms']:
                    d = bytes(f"{s}\0", "utf-8")
                    f.write(d)
                    d = struct.pack(">I", n)
                    f.write(d)

    return res

def write_stardict_zip(paths: StarDictPaths):
    with ZipFile(paths['zip_path'], 'w') as z:
        a = [paths['ifo_path'],
             paths['idx_path'],
             paths['dic_path'],
             paths['syn_path'],
             paths['icon_path']]
        for p in a:
            if p is not None:
                # NOTE .parent to create a top level folder in .zip
                z.write(p, p.relative_to(paths['unzipped_dir'].parent))

def export_words_as_stardict_zip(words: List[DictEntry],
                                 ifo: StarDictIfo,
                                 zip_path: Path,
                                 icon_path: Optional[Path] = None):

    name = zip_path.name.replace('.zip', '')
    # No spaces in the filename and dict files.
    name = name.replace(' ', '-')

    # NOTE: A toplevel folder is created in the zip file, with the file name of the .zip file.
    # E.g. ncped.zip will contain ncped/ncped.ifo
    unzipped_dir: Path = SIMSAPA_DIR.joinpath("new_stardict").joinpath(name)
    if unzipped_dir.exists():
        shutil.rmtree(unzipped_dir)
    unzipped_dir.mkdir(parents=True)

    zip_icon_path = None

    if icon_path is not None and icon_path.exists():
        ext = icon_path.suffix
        zip_icon_path = unzipped_dir.joinpath(f"{name}{ext}")
        shutil.copy(icon_path, zip_icon_path)

    paths = StarDictPaths(
        zip_path = zip_path,
        unzipped_dir = unzipped_dir,
        icon_path = zip_icon_path,
        ifo_path = unzipped_dir.joinpath(f"{name}.ifo"),
        idx_path = unzipped_dir.joinpath(f"{name}.idx"),
        dic_path = unzipped_dir.joinpath(f"{name}.dict.dz"),
        syn_path = unzipped_dir.joinpath(f"{name}.syn"),
    )

    ifo['version'] = '3.0.0'
    ifo['wordcount'] = f"{len(words)}"
    ifo['sametypesequence'] = 'h'
    ifo['date'] = datetime.datetime.utcnow().replace(microsecond=0).isoformat()

    res = write_words(words, paths)

    ifo['idxoffsetbits'] = "32"
    ifo['idxfilesize'] = f"{res['idx_size']}"
    ifo['synwordcount'] = f"{res['syn_count']}"

    write_ifo(ifo, paths)

    write_stardict_zip(paths)

    if unzipped_dir.exists():
        shutil.rmtree(unzipped_dir)



import logging as _logging

from sqlalchemy import (MetaData, Table, Column, Integer, String,
                        ForeignKey, Boolean, DateTime, LargeBinary)

from sqlalchemy import func
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

# -------------------------------------------------------------------------
# NOTE: The schema label identifies the attached database. This is the only
# difference between appdata_models.py and userdata_models.py.
# -------------------------------------------------------------------------

metadata = MetaData(schema='appdata')
Base = declarative_base(metadata=metadata)

logger = _logging.getLogger(__name__)

assoc_sutta_authors = Table(
    'sutta_authors',
    Base.metadata,
    Column('sutta_id', Integer, ForeignKey("suttas.id", ondelete="CASCADE"), primary_key=True, nullable=False),
    Column('author_id', Integer, ForeignKey("authors.id", ondelete="CASCADE"), primary_key=True, nullable=False),
)

assoc_document_tags = Table(
    'document_tags',
    Base.metadata,
    Column('document_id', Integer, ForeignKey("documents.id", ondelete="CASCADE"), primary_key=True, nullable=False),
    Column('tag_id', Integer, ForeignKey("tags.id", ondelete="CASCADE"), primary_key=True, nullable=False),
)

assoc_memo_tags = Table(
    'memo_tags',
    Base.metadata,
    Column('memo_id', Integer, ForeignKey("memos.id", ondelete="CASCADE"), primary_key=True, nullable=False),
    Column('tag_id', Integer, ForeignKey("tags.id", ondelete="CASCADE"), primary_key=True, nullable=False),
)


class Author(Base):
    __tablename__ = "authors"

    id = Column(Integer, primary_key=True)
    uid = Column(String, nullable=False, unique=True) # sujato
    full_name = Column(String) # Sujato Bhikkhu
    description = Column(String) # Translated for SuttaCentral by Sujato Bhikkhu

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())

    suttas = relationship("Sutta", secondary=assoc_sutta_authors, back_populates="authors")


class Sutta(Base):
    __tablename__ = "suttas"

    id = Column(Integer, primary_key=True)
    uid = Column(String, nullable=False, unique=True) # dn1/pli/ms
    group_path = Column(String) # /sutta-pitaka/digha-nikaya/silakkhandha-vagga
    group_index = Column(Integer) # 1
    sutta_ref = Column(String) # DN 1
    sutta_ref_pts = Column(String) # DN i 1
    language = Column(String) # pli / en
    order_index = Column(Integer)

    # --- Content props ---
    title = Column(String) # Brahmajāla: The Root of All Things
    title_pali = Column(String) # Brahmajāla
    title_trans = Column(String) # The Root of All Things
    description = Column(String)
    content_plain = Column(String) # content in plain text
    content_html = Column(String) # content in HTML

    # --- Source ---
    source_info = Column(String)
    source_language = Column(String)
    message = Column(String)
    copyright = Column(String)
    license = Column(String)

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())

    authors = relationship("Author", secondary=assoc_sutta_authors, back_populates="suttas")


class Dictionary(Base):
    __tablename__ = "dictionaries"

    id = Column(Integer, primary_key=True)
    label = Column(String, nullable=False, unique=True)
    title = Column(String, nullable=False)

    creator = Column(String)
    description = Column(String)
    feedback_email = Column(String)
    feedback_url = Column(String)
    version = Column(String)
    data_zip_url = Column(String)
    info_json_url = Column(String)
    url_synced_at = Column(String)
    has_update = Column(Boolean)

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())

    dict_words = relationship("DictWord", back_populates="dictionary", passive_deletes=True)


class DictWord(Base):
    __tablename__ = "dict_words"

    id = Column(Integer, primary_key=True)
    dictionary_id = Column(Integer, ForeignKey("dictionaries.id", ondelete="CASCADE"), nullable=False)
    uid = Column(String, nullable=False, unique=True)
    word = Column(String, nullable=False)
    word_nom_sg = Column(String)
    inflections = Column(String)
    phonetic = Column(String)
    transliteration = Column(String)

    # --- Meaning ---
    meaning_order = Column(Integer)
    definition_plain = Column(String)
    definition_html = Column(String)
    summary = Column(String)

    # --- Associated words ---
    synonyms = Column(String)
    antonyms = Column(String)
    homonyms = Column(String)
    also_written_as = Column(String)
    see_also = Column(String)

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())

    dictionary = relationship("Dictionary", back_populates="dict_words")
    examples = relationship("Example", back_populates="dict_word", passive_deletes=True)


class Example(Base):
    __tablename__ = "examples"

    id = Column(Integer, primary_key=True)
    dict_word_id = Column(Integer, ForeignKey("dict_words.id", ondelete="CASCADE"), nullable=False)

    source_ref = Column(String)
    source_title = Column(String)

    text_html = Column(String)
    translation_html = Column(String)

    highlight = Column(String)

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())

    dict_word = relationship("DictWord", back_populates="examples")


class Document(Base):
    __tablename__ = "documents"

    id = Column(Integer, primary_key=True)
    filepath = Column(String, nullable=False, unique=True)
    title = Column(String)
    author = Column(String)

    # --- Cover ---
    cover_data = Column(LargeBinary)
    cover_width = Column(Integer)
    cover_height = Column(Integer)
    cover_stride = Column(Integer)

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())

    tags = relationship("Tag", secondary=assoc_document_tags, back_populates="documents")


class Deck(Base):
    __tablename__ = "decks"

    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False, unique=True)

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())

    memos = relationship("Memo", back_populates="deck")


class Memo(Base):
    __tablename__ = "memos"

    id = Column(Integer, primary_key=True)
    deck_id = Column(Integer, ForeignKey("decks.id", ondelete="NO ACTION"))

    # --- Content ---
    fields_json = Column(String) # Front, Back

    anki_model_name = Column(String) # Basic, Cloze
    anki_note_id = Column(Integer)
    anki_synced_at = Column(DateTime)

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())

    deck = relationship("Deck", back_populates="memos")
    tags = relationship("Tag", secondary=assoc_memo_tags, back_populates="memos")


class MemoAssociation(Base):
    __tablename__ = "memo_associations"

    id = Column(Integer, primary_key=True)
    memo_id = Column(Integer, ForeignKey("memos.id", ondelete="CASCADE"), nullable=False)
    associated_table = Column(String, nullable=False)
    associated_id = Column(Integer, nullable=False)

    page_number = Column(Integer)
    location = Column(String)


class Annotation(Base):
    __tablename__ = "annotations"

    id = Column(Integer, primary_key=True)
    ann_type = Column(String)
    text = Column(String)

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())


class AnnotationAssociation(Base):
    __tablename__ = "annotation_associations"

    id = Column(Integer, primary_key=True)
    annotation_id = Column(Integer, ForeignKey("annotations.id", ondelete="CASCADE"))
    associated_table = Column(String, nullable=False)
    associated_id = Column(Integer, nullable=False)

    page_number = Column(Integer)
    location = Column(Integer)


class Tag(Base):
    __tablename__ = "tags"

    id = Column(Integer, primary_key=True)
    name = Column(String)

    documents = relationship("Document", secondary=assoc_document_tags, back_populates="tags")
    memos = relationship("Memo", secondary=assoc_memo_tags, back_populates="tags")


class Link(Base):
    __tablename__ = "links"

    id = Column(Integer, primary_key=True)
    label = Column(String)
    from_table = Column(String, nullable=False)
    from_id = Column(Integer, nullable=False)
    from_page_number = Column(Integer)
    to_table = Column(String, nullable=False)
    to_id = Column(Integer, nullable=False)
    to_page_number = Column(Integer)


class AppSetting(Base):
    __tablename__ = "app_settings"

    id = Column(Integer, primary_key=True)
    key = Column(String, nullable=False, unique=True)
    value = Column(String)

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())



import logging as _logging
import os.path
from functools import partial
from typing import List, Optional
from sqlalchemy.sql import func

from PyQt5.QtCore import QAbstractListModel, Qt
from PyQt5.QtGui import QIcon, QImage, QPixmap
from PyQt5.QtWidgets import (QFileDialog, QLabel, QMainWindow,
                             QMessageBox)
from simsapa.assets import icons_rc  # noqa: F401

from ..app.file_doc import FileDoc
from ..app.db import appdata_models as Am
from ..app.db import userdata_models as Um

from ..app.types import AppData, UDocument
from ..assets.ui.library_browser_window_ui import Ui_LibraryBrowserWindow

logger = _logging.getLogger(__name__)


class DocumentListModel(QAbstractListModel):
    def __init__(self, *args, docs=None, **kwargs):
        super(DocumentListModel, self).__init__(*args, **kwargs)
        self.docs = docs or []

    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole:
            filepath = self.docs[index.row()].filepath
            name = os.path.basename(filepath)
            return name

        if role == Qt.ItemDataRole.DecorationRole:
            filepath = self.docs[index.row()].filepath
            if not os.path.exists(filepath):
                return QIcon(":close")

    def rowCount(self, index):
        return len(self.docs)


class LibraryBrowserWindow(QMainWindow, Ui_LibraryBrowserWindow):
    def __init__(self, app_data: AppData, parent=None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self._app_data: AppData = app_data

        self.model = DocumentListModel(docs=self._get_all_documents())

        self.documents_list.setModel(self.model)
        self.sel_model = self.documents_list.selectionModel()

        self._ui_setup()
        self._connect_signals()

        self.statusbar.showMessage("Ready", 3000)

    def _ui_setup(self):
        self.status_msg = QLabel("")
        self.statusbar.addPermanentWidget(self.status_msg)

        self.search_input.setFocus()
        self._show_document_clear()

    def _get_all_documents(self) -> List[UDocument]:
        results = []

        res = self._app_data.db_session.query(Am.Document).all()
        results.extend(res)
        res = self._app_data.db_session.query(Um.Document).all()
        results.extend(res)

        return results

    def _handle_query(self):
        pass

    def get_selected_document(self) -> Optional[UDocument]:
        a = self.documents_list.selectedIndexes()
        if not a:
            return None

        item = a[0]
        return self.model.docs[item.row()]

    def remove_selected_document(self):
        a = self.documents_list.selectedIndexes()
        if not a:
            return None

        # Remove from model
        item = a[0]
        doc = self.model.docs[item.row()]
        doc_id = doc.id
        schema = doc.metadata.schema

        del self.model.docs[item.row()]
        self.model.layoutChanged.emit()
        self.documents_list.clearSelection()
        self._show_document_clear()

        # Remove from database

        if schema == 'appdata':
            db_item = self._app_data.db_session \
                                    .query(Am.Document) \
                                    .filter(Am.Document.id == doc_id) \
                                    .first()
        else:
            db_item = self._app_data.db_session \
                                    .query(Um.Document) \
                                    .filter(Um.Document.id == doc_id) \
                                    .first()

        self._app_data.db_session.delete(db_item)
        self._app_data.db_session.commit()

    def _handle_document_select(self):
        doc = self.get_selected_document()
        if doc:
            self._show_document(doc)

    def _show_document_clear(self):
        self.status_msg.clear()
        self.doc_title.clear()
        self.doc_author.clear()
        self.doc_cover.clear()

    def _show_document(self, doc: UDocument):
        self.status_msg.setText(doc.filepath)
        self.doc_title.setText(doc.title)
        self.doc_author.setText(doc.author)

        if doc.cover_data:
            img = QImage(
                doc.cover_data,
                doc.cover_width,
                doc.cover_height,
                doc.cover_stride,
                QImage.Format.Format_RGB888
            )
            self.doc_cover.setPixmap(QPixmap.fromImage(img))

    def _documents_search_query(self, query: str):
        results = []

        res = self._app_data.db_session \
                            .query(Am.Document) \
                            .filter(Am.Document.filepath.like(f"%{query}%")) \
                            .all()
        results.extend(res)

        res = self._app_data.db_session \
                            .query(Um.Document) \
                            .filter(Um.Document.filepath.like(f"%{query}%")) \
                            .all()
        results.extend(res)

        return results

    def add_document(self, filepath):
        if not os.path.exists(filepath):
            logger.error(f"File doesn't exist: {filepath}")
            return

        file_doc = FileDoc(filepath)

        title = file_doc.title or "Unknown"

        author = file_doc.author or "Unknown"

        page_image = file_doc.page_image(0, 0.7)

        # Insert or update database record

        item = self._app_data.db_session \
                             .query(Um.Document) \
                             .filter(Um.Document.filepath == filepath) \
                             .first()
        if item is None:
            logger.info(f"Add new: {filepath}")

            if page_image:
                db_doc = Um.Document(
                    filepath=filepath,
                    title=title,
                    author=author,
                    cover_data=page_image.image_bytes,
                    cover_width=page_image.width,
                    cover_height=page_image.height,
                    cover_stride=page_image.stride,
                    created_at=func.now(),
                )
            else:
                db_doc = Um.Document(
                    filepath=filepath,
                    title=title,
                    author=author,
                    created_at=func.now(),
                )

            try:
                self._app_data.db_session.add(db_doc)
                self._app_data.db_session.commit()

                # Add to model
                self.model.docs.append(db_doc)

            except Exception as e:
                logger.error(e)

        else:
            logger.info(f"Update: {item.filepath}")

            if page_image:
                values = {
                    'filepath': filepath,
                    'title': title,
                    'author': author,
                    'cover_data': page_image.image_bytes,
                    'cover_width': page_image.width,
                    'cover_height': page_image.height,
                    'cover_stride': page_image.stride,
                    'updated_at': func.now(),
                }
            else:
                values = {
                    'filepath': filepath,
                    'title': title,
                    'author': author,
                    'updated_at': func.now(),
                }

            try:
                self._app_data.db_session \
                              .query(Um.Document) \
                              .filter(Um.Document.id == item.id) \
                              .update(values)

                self._app_data.db_session.commit()

            except Exception as e:
                logger.error(e)

        self.model.layoutChanged.emit()

    def _add_document_dialog(self):
        file_paths, _ = QFileDialog.getOpenFileNames(
            None,
            "Add Documents to the Library...",
            "",
            "PDF or Epub Files (*.pdf *.epub)")

        for p in file_paths:
            self.add_document(p)

    def _remove_document_dialog(self):
        doc = self.get_selected_document()
        if not doc:
            return

        reply = QMessageBox.question(self,
                                     'Remove Document...',
                                     'Remove this item from the Library? (Files are NOT deleted.)',
                                     QMessageBox.Yes | QMessageBox.No,
                                     QMessageBox.No)
        if reply == QMessageBox.Yes:
            self.remove_selected_document()

    def _connect_signals(self):
        self.action_Close_Window \
            .triggered.connect(partial(self.close))

        self.action_Add \
            .triggered.connect(partial(self._add_document_dialog))

        self.action_Remove \
            .triggered.connect(partial(self._remove_document_dialog))

        self.search_button.clicked.connect(partial(self._handle_query))
        self.search_input.textChanged.connect(partial(self._handle_query))

        self.sel_model.selectionChanged.connect(partial(self._handle_document_select))



import logging as _logging
import json
import requests
from functools import partial
from typing import List, Optional

from PyQt5.QtCore import QAbstractListModel, Qt, QItemSelectionModel
from PyQt5.QtWidgets import (QLabel, QMainWindow,  QMessageBox)
from sqlalchemy.sql import func

from simsapa.assets import icons_rc  # noqa: F401

from ..app.db import appdata_models as Am
from ..app.db import userdata_models as Um

from ..app.types import AppData, UMemo
from ..assets.ui.memos_browser_window_ui import Ui_MemosBrowserWindow

logger = _logging.getLogger(__name__)


class MemoListModel(QAbstractListModel):
    def __init__(self, *args, memos=None, **kwargs):
        super(MemoListModel, self).__init__(*args, **kwargs)
        self.memos = memos or []

    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole:
            fields = json.loads(self.memos[index.row()].fields_json)
            text = " ".join(fields.values())
            return text

    def rowCount(self, index):
        if self.memos:
            return len(self.memos)
        else:
            return 0


class MemosBrowserWindow(QMainWindow, Ui_MemosBrowserWindow):
    def __init__(self, app_data: AppData, parent=None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self._app_data: AppData = app_data

        self.model = MemoListModel(memos=self._get_all_memos())

        self.memos_list.setModel(self.model)
        self.sel_model = self.memos_list.selectionModel()

        self._ui_setup()
        self._connect_signals()

        self.statusbar.showMessage("Ready", 3000)

    def _ui_setup(self):
        self.status_msg = QLabel("")
        self.statusbar.addPermanentWidget(self.status_msg)

        self.front_input.setFocus()
        self._show_memo_clear()

    def _get_all_memos(self) -> List[UMemo]:
        results = []
        res = self._app_data.db_session.query(Am.Memo).all()
        results.extend(res)
        res = self._app_data.db_session.query(Um.Memo).all()
        results.extend(res)

        return results

    def _handle_query(self):
        pass

    def get_selected_memo(self) -> Optional[UMemo]:
        a = self.memos_list.selectedIndexes()
        if not a:
            return None

        item = a[0]
        return self.model.memos[item.row()]

    def remove_selected_memo(self):
        a = self.memos_list.selectedIndexes()
        if not a:
            return None

        # Remove from model
        item = a[0]
        memo = self.model.memos[item.row()]
        memo_id = memo.id
        schema = memo.metadata.schema

        del self.model.memos[item.row()]
        self.model.layoutChanged.emit()
        self.memos_list.clearSelection()
        self._show_memo_clear()

        # Remove from database

        if schema == 'appdata':
            db_item = self._app_data.db_session \
                                    .query(Am.Memo) \
                                    .filter(Am.Memo.id == memo_id) \
                                    .first()
        else:
            db_item = self._app_data.db_session \
                                    .query(Um.Memo) \
                                    .filter(Um.Memo.id == memo_id) \
                                    .first()

        self._app_data.db_session.delete(db_item)
        self._app_data.db_session.commit()

    def _handle_memo_select(self):
        memo = self.get_selected_memo()
        if memo:
            self._show_memo(memo)

    def _show_memo_clear(self):
        self.status_msg.clear()
        self.front_input.clear()
        self.back_input.clear()

    def _show_memo(self, memo: UMemo):
        fields = json.loads(memo.fields_json) # type: ignore
        self.front_input.setPlainText(fields['Front'])
        self.back_input.setPlainText(fields['Back'])

    def clear_memo(self):
        self.sel_model.clearSelection()
        self.front_input.clear()
        self.back_input.clear()

    def _memos_search_query(self, query: str):
        return []

    def add_memo(self):
        self.sel_model.clearSelection()

        front = self.front_input.toPlainText()
        back = self.back_input.toPlainText()

        if len(front) == 0 and len(back) == 0:
            logger.info("Empty content, cancel adding.")
            return

        # Insert database record

        logger.info("Adding new memo")

        deck = self._app_data.db_session.query(Um.Deck).first()
        if deck is None:
            logger.error("Can't find Deck")
            return

        memo_fields = {
            'Front': front,
            'Back': back
        }

        memo = Um.Memo(
            deck_id=deck.id,
            fields_json=json.dumps(memo_fields),
            created_at=func.now(),
        )

        try:
            self._app_data.db_session.add(memo)
            self._app_data.db_session.commit()

            # Add to model
            if self.model.memos:
                self.model.memos.append(memo)
            else:
                self.model.memos = [memo]

        except Exception as e:
            logger.error(e)

        index = self.model.index(len(self.model.memos) - 1)
        self.memos_list.selectionModel().select(index, QItemSelectionModel.SelectionFlag.Select)

        self.model.layoutChanged.emit()

    def update_selected_memo_fields(self):
        memo = self.get_selected_memo()
        if memo is None:
            return

        fields = {
            'Front': self.front_input.toPlainText(),
            'Back': self.back_input.toPlainText()
        }

        memo.fields_json = json.dumps(fields) # type: ignore

        self._app_data.db_session.commit()
        self.model.layoutChanged.emit()

    def remove_memo_dialog(self):
        memo = self.get_selected_memo()
        if not memo:
            return

        reply = QMessageBox.question(self,
                                     'Remove Memo...',
                                     'Remove this item?',
                                     QMessageBox.Yes | QMessageBox.No,
                                     QMessageBox.No)
        if reply == QMessageBox.Yes:
            self.remove_selected_memo()

    def test_anki_live_or_notify(self) -> bool:
        if not is_anki_live():
            QMessageBox.information(self,
                                    "Anki is not connected",
                                    "Anki must be running with the AnkiConnect plugin.",
                                    QMessageBox.Ok)
            return False
        return True

    def sync_to_anki(self):
        if not self.test_anki_live_or_notify():
            return

        res = anki_invoke('deckNames')
        if 'Simsapa' not in res:
            anki_invoke('createDeck', deck='Simsapa')

        for memo in self.model.memos:

            fields = json.loads(memo.fields_json)

            if memo.anki_note_id:
                logger.info("Updating Anki note...")

                anki_invoke(
                    'updateNoteFields',
                    note={
                        'id': memo.anki_note_id,
                        'fields': {
                            'Front': fields['Front'],
                            'Back': fields['Back'],
                        },
                    }
                )

            else:
                logger.info("Creating Anki note...")

                res = anki_invoke(
                    'addNote',
                    note={
                        'deckName': 'Simsapa',
                        'modelName': 'Basic',
                        'fields': {
                            'Front': fields['Front'],
                            'Back': fields['Back'],
                        },
                        'options': {
                            'allowDuplicate': True,
                            'duplicateScope': 'deck',
                            'duplicateScopeOptions': {
                                'deckName': 'Simsapa',
                                'checkChildren': False
                            }
                        },
                    }
                )

                memo.anki_note_id = res
                self._app_data.db_session.commit()

    def _connect_signals(self):
        self.action_Close_Window \
            .triggered.connect(partial(self.close))

        self.action_Add \
            .triggered.connect(partial(self.add_memo))

        self.action_Clear \
            .triggered.connect(partial(self.clear_memo))

        self.action_Remove \
            .triggered.connect(partial(self.remove_memo_dialog))

        self.action_Sync_to_Anki \
            .triggered.connect(partial(self.sync_to_anki))

        self.search_button.clicked.connect(partial(self._handle_query))
        self.search_input.textChanged.connect(partial(self._handle_query))

        self.front_input.textChanged.connect(partial(self.update_selected_memo_fields))
        self.back_input.textChanged.connect(partial(self.update_selected_memo_fields))

        self.sel_model.selectionChanged.connect(partial(self._handle_memo_select))


def is_anki_live() -> bool:
    try:
        res = requests.get('http://localhost:8765')
    except Exception:
        return False

    return res.status_code == 200


def anki_request(action, **params):
    return {'action': action, 'params': params, 'version': 6}


def anki_invoke(action, **params):
    res = requests.get(url='http://localhost:8765', json=anki_request(action, **params))
    response = res.json()

    if len(response) != 2:
        raise Exception('response has an unexpected number of fields')

    if 'error' not in response:
        raise Exception('response is missing required error field')

    if 'result' not in response:
        raise Exception('response is missing required result field')

    if response['error'] is not None:
        raise Exception(response['error'])

    return response['result']



import os.path
import logging as _logging
import shutil
from typing import List, Optional, TypedDict

from pathlib import Path

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QComboBox, QDialog, QFileDialog, QLabel, QLineEdit, QTabWidget

from simsapa.app.types import AppData

from ..assets.ui.import_stardict_dialog_ui import Ui_ImportStarDictDialog

from ..app.db import userdata_models as Um

from ..app.stardict import StarDictPaths, StarDictIfo, parse_stardict_zip, parse_ifo
from ..app.db.stardict import import_stardict_into_db_as_new, import_stardict_into_db_update_existing

logger = _logging.getLogger(__name__)

class DictData(TypedDict):
    id: int
    select_idx: int
    title: str
    label: str

class ImportStarDictDialog(QDialog, Ui_ImportStarDictDialog):

    accepted = pyqtSignal(dict) # type: ignore
    stardict_paths: Optional[StarDictPaths] = None
    stardict_ifo: Optional[StarDictIfo] = None

    msg: QLabel
    title: QLineEdit
    label: QLineEdit
    select_title: QComboBox
    tabWidget: QTabWidget

    def __init__(self, app_data: AppData, parent=None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self._app_data: AppData = app_data

        self.dict_select_data: List[DictData] = []

        self._ui_setup()
        self._connect_signals()

    def _ui_setup(self):
        self.setWindowFlags(Qt.WindowType.WindowStaysOnTopHint)

        self.title.setFixedSize(500, 40)
        self.label.setFixedSize(500, 40)

        self.info.setText("Select a StarDict .zip file to open.")
        self.msg.clear()
        self.selected_info.clear()

        # fill select_title with title - label
        self.select_title.insertItem(0, 'Select dictionary')

        # TODO also consider appdata
        a = self._app_data.db_session \
                          .query(Um.Dictionary.id,
                                 Um.Dictionary.title,
                                 Um.Dictionary.label) \
                          .all()

        # Add the titles to the selection, with the (dictionary.id, dictionary.label) as item data.
        select_item_counter = 1
        for i in a:
            self.select_title.insertItem(
                select_item_counter,
                f"{i[1]} - {i[2]}",
                (i[0], i[2]))

            self.dict_select_data.append(DictData(
                id = i[0],
                select_idx = select_item_counter,
                title = i[1],
                label = i[2],
            ))

            select_item_counter += 1

    def _connect_signals(self):
        self.import_new_btn.clicked.connect(self.import_new_pressed)
        self.update_existing_btn.clicked.connect(self.update_existing_pressed)
        self.close_btn.clicked.connect(self.close_pressed)
        self.open_btn.clicked.connect(self.open_file_dialog)

        self.title.textChanged.connect(self.unlock)
        self.label.textChanged.connect(self.unlock)
        self.select_title.currentIndexChanged.connect(self.unlock)

    def import_new_pressed(self):
        values = {
            'action': 'import_new',
            'stardict_paths': self.stardict_paths,
            'title': self.title.text(),
            'label': self.label.text(),
        }
        self.accepted.emit(values)
        self.accept()

    def update_existing_pressed(self):
        idx = self.select_title.currentIndex()
        if idx == 0:
            return
        dictionary_id, label = self.select_title.itemData(idx)

        values = {
            'action': 'update_existing',
            'stardict_paths': self.stardict_paths,
            'dictionary_id': dictionary_id,
            'label': label,
        }
        self.accepted.emit(values)
        self.accept()

    def close_pressed(self):
        self.close()

    def open_file_dialog(self):
        file_path, _ = QFileDialog.getOpenFileName(
            None,
            "Open File...",
            "",
            "ZIP Files (*.zip)")

        if len(file_path) != 0:
            self.open_stardict_zip(file_path)
            self.unlock()

    def open_stardict_zip(self, zip_path: str):
        self.msg.clear()

        try:
            self.stardict_paths = parse_stardict_zip(Path(zip_path))
            self.stardict_ifo = parse_ifo(self.stardict_paths)
        except Exception as e:
            logger.error(e)
            self.msg.setText(f"{e}")
            return

        a = list(filter(lambda x: len(x) > 0,
                   [os.path.basename(self.stardict_paths['zip_path']),
                    self.stardict_ifo['bookname'],
                    self.stardict_ifo['description'],
                    self.stardict_ifo['wordcount'] + " words",
                    self.stardict_ifo['date'],
                    self.stardict_ifo['author'],
                   ]))

        self.info.setText("\n".join(a))

        title = self.stardict_ifo['bookname']

        self.title.setText(title)

        a = list(filter(lambda x: x['title'] == title, self.dict_select_data))

        # If a dictionary with this title exists, select the update tab
        # with the existing dictionary selected.
        if len(a) > 0:
            self.tabWidget.setCurrentIndex(1)
            self.select_title.setCurrentIndex(a[0]['select_idx'])


    def unlock(self):
        if self.stardict_ifo is None:
            return

        self.unlock_import_new()
        self.unlock_update_existing()

    def unlock_import_new(self):
        self.import_new_btn.setDisabled(True)

        if len(self.title.text()) < 3:
            self.msg.setText("Title must be at least 3 characters long.")
            return

        if len(self.label.text()) < 3:
            self.msg.setText("Label must be at least 3 characters long.")
            # TODO label must be unique
            return

        self.msg.clear()
        self.import_new_btn.setEnabled(True)

    def unlock_update_existing(self):
        self.update_existing_btn.setDisabled(True)

        idx = self.select_title.currentIndex()
        if idx == 0:
            return

        self.update_existing_btn.setEnabled(True)

    def do_import(self, values):
        paths = values['stardict_paths']
        action = values['action']
        label = values['label']

        if action == 'import_new':
            import_stardict_into_db_as_new(self._app_data.db_session, 'userdata', paths, label)
        elif action == 'update_existing':
            id = values['dictionary_id']
            import_stardict_into_db_update_existing(self._app_data.db_session, 'userdata', paths, id, label)

        # remove zip extract dir
        if paths['unzipped_dir'].exists():
            shutil.rmtree(paths['unzipped_dir'])

class HasImportStarDictDialog():
    _app_data: AppData

    def init_stardict_import_dialog(self):
        pass

    def show_import_from_stardict_dialog(self):
        d = ImportStarDictDialog(self._app_data)
        d.accepted.connect(d.do_import)
        d.exec_()



import os
import logging as _logging
import json
from pathlib import Path
import queue

from functools import partial
from typing import List, Optional

from PyQt5.QtCore import QUrl, QTimer
from PyQt5.QtGui import QCloseEvent
from PyQt5.QtWidgets import (QLabel, QMainWindow, QListWidgetItem,
                             QHBoxLayout, QPushButton, QSizePolicy, QAction, QMessageBox,
                             QComboBox)
from PyQt5.QtWebEngineWidgets import QWebEngineView

from sqlalchemy import or_

from simsapa import ASSETS_DIR, APP_QUEUES
from ..app.db import appdata_models as Am
from ..app.db import userdata_models as Um
from ..app.db.search import SearchResult
from ..app.types import AppData, USutta, UDictWord, UDocument
from ..app.graph import generate_graph, all_nodes_and_edges
from ..assets.ui.links_browser_window_ui import Ui_LinksBrowserWindow
from .search_item import SearchItemWidget

logger = _logging.getLogger(__name__)


class LinksBrowserWindow(QMainWindow, Ui_LinksBrowserWindow):

    def __init__(self, app_data: AppData, parent=None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self.link_table: QComboBox;

        self.features = []
        self._app_data: AppData = app_data
        self._results: List[SearchResult] = []

        self._current_from: Optional[SearchResult] = None
        self._current_to: Optional[SearchResult] = None

        self.queue_id = 'window_' + str(len(APP_QUEUES))
        APP_QUEUES[self.queue_id] = queue.Queue()
        self.messages_url = f'{self._app_data.api_url}/queues/{self.queue_id}'

        self.graph_path: Path = ASSETS_DIR.joinpath(f"{self.queue_id}.html")

        self.timer = QTimer()
        self.timer.timeout.connect(self.handle_messages)
        self.timer.start(300)

        self._ui_setup()
        self._connect_signals()

        self.statusbar.showMessage("Ready", 3000)

    def closeEvent(self, event: QCloseEvent):
        if self.queue_id in APP_QUEUES.keys():
            del APP_QUEUES[self.queue_id]

        if self.graph_path.exists():
            self.graph_path.unlink()

        event.accept()

    def handle_messages(self):
        if self.queue_id in APP_QUEUES.keys():
            try:
                s = APP_QUEUES[self.queue_id].get_nowait()
                data = json.loads(s)
                if data['action'] == 'show_sutta':
                    self._show_sutta_from_message(data['arg'])

                APP_QUEUES[self.queue_id].task_done()
            except queue.Empty:
                pass

    def _ui_setup(self):
        self.status_msg = QLabel("")
        self.statusbar.addPermanentWidget(self.status_msg)

        self._setup_pali_buttons()

        self.setup_content_graph()
        self.show_network_graph()

        self.search_input.setFocus()

    def _setup_pali_buttons(self):
        lowercase = 'ā ī ū ṃ ṁ ṅ ñ ṭ ḍ ṇ ḷ'.split(' ')
        uppercase = 'Ā Ī Ū Ṃ Ṁ Ṅ Ñ Ṭ Ḍ Ṇ Ḷ'.split(' ')

        lowercase_row = QHBoxLayout()

        for i in lowercase:
            btn = QPushButton(i)
            btn.setFixedSize(30, 30)
            btn.clicked.connect(partial(self._append_to_query, i))
            lowercase_row.addWidget(btn)

        uppercase_row = QHBoxLayout()

        for i in uppercase:
            btn = QPushButton(i)
            btn.setFixedSize(30, 30)
            btn.clicked.connect(partial(self._append_to_query, i))
            uppercase_row.addWidget(btn)

        self.pali_buttons_layout.addLayout(lowercase_row)
        self.pali_buttons_layout.addLayout(uppercase_row)

    def setup_content_graph(self):
        self.content_graph = QWebEngineView()
        self.content_graph.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.content_graph.setHtml('')
        self.content_graph.show()
        self.links_layout.addWidget(self.content_graph)

    def show_network_graph(self):
        (nodes, edges) = all_nodes_and_edges(app_data=self._app_data)
        generate_graph(nodes, edges, [], self.queue_id, self.graph_path, self.messages_url)

        self.content_graph.load(QUrl('file://' + str(self.graph_path.absolute())))

    def _append_to_query(self, s: str):
        a = self.search_input.text()
        self.search_input.setText(a + s)
        self.search_input.setFocus()

    def _handle_table_changed(self):
        idx = self.link_table.currentIndex()
        link_table = self.link_table.itemText(idx)
        if link_table == 'Suttas' or link_table == 'DictWords':
            self.set_page_number.setEnabled(False)
            self.page_number.setEnabled(False)
        else:
            self.set_page_number.setEnabled(True)
            self.page_number.setEnabled(True)

        self._handle_query()

    def _handle_query(self, min_length=4):
        query = self.search_input.text()
        if len(query) < min_length:
            return

        idx = self.link_table.currentIndex()
        link_table = self.link_table.itemText(idx)

        if link_table == 'Suttas':
            self._results = self._sutta_search_query(query)
        elif link_table == 'DictWords':
            self._results = self._dict_words_search_query(query)
        else:
            self._results = self._documents_search_query(query)

        self.results_list.clear()

        # FIXME paginate results. Too many results hang the UI while rendering.
        self._results = self._results[0:100]

        for x in self._results:
            w = SearchItemWidget()
            w.setTitle(x['title'])

            w.setSnippet(x['snippet'])

            item = QListWidgetItem(self.results_list)
            item.setSizeHint(w.sizeHint())

            self.results_list.addItem(item)
            self.results_list.setItemWidget(item, w)

    def _sutta_search_query(self, query: str) -> List[SearchResult]:
        db_results: List[USutta] = []

        r = self._app_data.db_session \
            .query(Am.Sutta) \
            .filter(or_(
                Am.Sutta.content_plain.like(f"%{query}%"),
                Am.Sutta.content_html.like(f"%{query}%"))) \
            .all()
        db_results.extend(r)

        r = self._app_data.db_session \
            .query(Um.Sutta) \
            .filter(or_(
                Um.Sutta.content_plain.like(f"%{query}%"),
                Um.Sutta.content_html.like(f"%{query}%"))) \
            .all()
        db_results.extend(r)

        def to_search_result(x: USutta):
            snippet = ''
            if x.content_html:
                snippet = x.content_html[0:400].strip()
            elif x.content_plain:
                snippet = x.content_plain[0:400].strip()

            return SearchResult(
                title=x.title.strip(),
                snippet=snippet,
                schema_name=x.metadata.schema,
                table_name=f"{x.metadata.schema}.suttas",
                id=x.id, # type: ignore
                page_number=None,
            )

        return list(map(to_search_result, db_results))

    def _dict_words_search_query(self, query: str) -> List[SearchResult]:
        db_results: List[UDictWord] = []

        r = self._app_data.db_session \
            .query(Am.DictWord) \
            .filter(Am.DictWord.word.like(f"%{query}%")) \
            .all()
        db_results.extend(r)

        r = self._app_data.db_session \
            .query(Um.DictWord) \
            .filter(Um.DictWord.word.like(f"%{query}%")) \
            .all()
        db_results.extend(r)

        def to_search_result(x: UDictWord):
            snippet = ''
            if x.definition_html:
                snippet = x.definition_html[0:400].strip()
            elif x.definition_plain:
                snippet = x.definition_plain[0:400].strip()

            return SearchResult(
                title=x.word.strip(),
                snippet=snippet,
                schema_name=x.metadata.schema,
                table_name=f"{x.metadata.schema}.dict_words",
                id=x.id, # type: ignore
                page_number=None,
            )

        return list(map(to_search_result, db_results))

    def _documents_search_query(self, query: str) -> List[SearchResult]:
        db_results: List[UDocument] = []

        r = self._app_data.db_session \
            .query(Am.Document) \
            .filter(or_(
                Am.Document.title.like(f"%{query}%"),
                Am.Document.filepath.like(f"%{query}%"))) \
            .all()
        db_results.extend(r)

        r = self._app_data.db_session \
            .query(Um.Document) \
            .filter(or_(
                Um.Document.title.like(f"%{query}%"),
                Um.Document.filepath.like(f"%{query}%"))) \
            .all()
        db_results.extend(r)

        def to_search_result(x: UDocument):
            title = x.title
            if x.author:
                title += ', ' + x.author

            return SearchResult(
                title=title.strip(),
                snippet='',
                schema_name=x.metadata.schema,
                table_name=f"{x.metadata.schema}.documents",
                id=x.id, # type: ignore
                page_number=None,
            )

        return list(map(to_search_result, db_results))

    def _show_sutta_from_message(self, info):
        sutta: Optional[USutta] = None

        if info['table'] == 'appdata.suttas':
            sutta = self._app_data.db_session \
                .query(Am.Sutta) \
                .filter(Am.Sutta.id == info['id']) \
                .first()

        elif info['table'] == 'userdata.suttas':
            sutta = self._app_data.db_session \
                .query(Um.Sutta) \
                .filter(Um.Sutta.id == info['id']) \
                .first()

        self._app_data.sutta_to_open = sutta
        self.action_Sutta_Search.activate(QAction.ActionEvent.Trigger)

    def _handle_set_from(self):
        selected_idx = self.results_list.currentRow()
        if selected_idx == -1:
            return
        if selected_idx < len(self._results):
            self._current_from = self._results[selected_idx]
        else:
            return

        if self.set_page_number.isEnabled() and self.set_page_number.isChecked():
            self._current_from['page_number'] = self.page_number.value()

        text = self._current_from['title']
        if self._current_from['page_number'] is not None:
            text += f"\npage {self._current_from['page_number']}"

        self.from_view.setPlainText(text)

    def _handle_set_to(self):
        selected_idx = self.results_list.currentRow()
        if selected_idx == -1:
            return
        if selected_idx < len(self._results):
            self._current_to = self._results[selected_idx]
        else:
            return

        if self.set_page_number.isEnabled() and self.set_page_number.isChecked():
            self._current_to['page_number'] = self.page_number.value()

        text = self._current_to['title']
        if self._current_to['page_number'] is not None:
            text += f"\npage {self._current_to['page_number']}"

        self.to_view.setPlainText(text)

    def _handle_create_link(self):
        if not self._check_from_and_to():
            return

        links = self._find_existing_links()

        if len(links) > 0:
            QMessageBox.information(self,
                                    "Link Exists",
                                    "This link already exists.",
                                    QMessageBox.Ok)
            return

        if self._current_from is None:
            return
        if self._current_to is None:
            return

        link = Um.Link(
            from_table=self._current_from['table_name'],
            from_id=self._current_from['id'],
            from_page_number=self._current_from['page_number'],
            to_table=self._current_to['table_name'],
            to_id=self._current_to['id'],
            to_page_number=self._current_to['page_number'],
        )

        try:
            self._app_data.db_session.add(link)
            self._app_data.db_session.commit()

            self.show_network_graph()
        except Exception as e:
            logger.error(e)

    def _handle_clear_link(self):
        self._current_from = None
        self._current_to = None
        self.from_view.setPlainText('')
        self.to_view.setPlainText('')

    def _check_from_and_to(self) -> bool:
        if not self._current_from or not self._current_to:
            QMessageBox.information(self,
                                    "Missing Information",
                                    "'From' and 'To' cannot be empty.",
                                    QMessageBox.Ok)
            return False

        keys = ['table', 'id', 'page_number']
        is_same = True
        for k in keys:
            if self._current_from[k] != self._current_to[k]:
                is_same = False

        if is_same:
            QMessageBox.information(self,
                                    "Self-Reference",
                                    "'From' and 'To' cannot be the same.",
                                    QMessageBox.Ok)
            return False

        return True

    def _find_existing_links(self):
        links = []

        if self._current_from is None:
            return []
        if self._current_to is None:
            return []

        r = self._app_data.db_session \
            .query(Am.Link) \
            .filter(Am.Link.from_table == self._current_from['table_name']) \
            .filter(Am.Link.from_id == self._current_from['id']) \
            .filter(Am.Link.from_page_number == self._current_from['page_number']) \
            .filter(Am.Link.to_table == self._current_to['table_name']) \
            .filter(Am.Link.to_id == self._current_to['id']) \
            .filter(Am.Link.to_page_number == self._current_to['page_number']) \
            .all()
        links.extend(r)

        r = self._app_data.db_session \
            .query(Am.Link) \
            .filter(Am.Link.to_table == self._current_from['table_name']) \
            .filter(Am.Link.to_id == self._current_from['id']) \
            .filter(Am.Link.to_page_number == self._current_from['page_number']) \
            .filter(Am.Link.from_table == self._current_to['table_name']) \
            .filter(Am.Link.from_id == self._current_to['id']) \
            .filter(Am.Link.from_page_number == self._current_to['page_number']) \
            .all()
        links.extend(r)

        r = self._app_data.db_session \
            .query(Um.Link) \
            .filter(Um.Link.from_table == self._current_from['table_name']) \
            .filter(Um.Link.from_id == self._current_from['id']) \
            .filter(Um.Link.from_page_number == self._current_from['page_number']) \
            .filter(Um.Link.to_table == self._current_to['table_name']) \
            .filter(Um.Link.to_id == self._current_to['id']) \
            .filter(Um.Link.to_page_number == self._current_to['page_number']) \
            .all()
        links.extend(r)

        r = self._app_data.db_session \
            .query(Um.Link) \
            .filter(Um.Link.to_table == self._current_from['table_name']) \
            .filter(Um.Link.to_id == self._current_from['id']) \
            .filter(Um.Link.to_page_number == self._current_from['page_number']) \
            .filter(Um.Link.from_table == self._current_to['table_name']) \
            .filter(Um.Link.from_id == self._current_to['id']) \
            .filter(Um.Link.from_page_number == self._current_to['page_number']) \
            .all()
        links.extend(r)

        return links

    def _handle_remove_link(self):
        if not self._check_from_and_to():
            return

        links = self._find_existing_links()

        if len(links) > 0:
            for x in links:
                try:
                    self._app_data.db_session.delete(x)
                    self._app_data.db_session.commit()

                except Exception as e:
                    logger.error(e)

            self.show_network_graph()
        else:
            QMessageBox.information(self,
                                    "No Link Found",
                                    "No link was found with these properties.",
                                    QMessageBox.Ok)

    def _connect_signals(self):
        self.action_Close_Window \
            .triggered.connect(partial(self.close))

        self.search_button.clicked.connect(partial(self._handle_query, min_length=1))
        self.search_input.textChanged.connect(partial(self._handle_query, min_length=4))
        self.link_table.currentIndexChanged.connect(partial(self._handle_table_changed))
        self.set_from_btn.clicked.connect(partial(self._handle_set_from))
        self.set_to_btn.clicked.connect(partial(self._handle_set_to))
        self.create_link_btn.clicked.connect(partial(self._handle_create_link))
        self.clear_link_btn.clicked.connect(partial(self._handle_clear_link))
        self.remove_link_btn.clicked.connect(partial(self._handle_remove_link))

        s = os.getenv('ENABLE_WIP_FEATURES')
        if s is not None and s.lower() == 'true':
            pass
        else:
            # don't show "Documents" option
            self.link_table.removeItem(2)
            # hide setting page number
            self.page_number.setVisible(False)
            self.set_page_number.setVisible(False)



from PyQt5.QtWidgets import QFrame, QHBoxLayout, QSpacerItem, QWidget, QVBoxLayout, QLabel, QSizePolicy

from simsapa.app.db.search import SearchResult

class SearchItemWidget(QWidget):
    def __init__(self, parent=None):
        super(SearchItemWidget, self).__init__(parent)

        self.layout: QVBoxLayout = QVBoxLayout()
        self.setLayout(self.layout)

        self.top_info: QHBoxLayout = QHBoxLayout()

        self.title = QLabel()
        self.author = QLabel()
        self.details = QLabel()

        self.top_spacer = QSpacerItem(20, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.top_info.addWidget(self.title)

        self.top_info.addItem(self.top_spacer)

        self.top_info.addWidget(self.author)
        self.top_info.addWidget(self.details)

        self.layout.addLayout(self.top_info)

        self.snippet = QLabel()
        self.snippet.setWordWrap(True)

        self.snippet.setFixedHeight(35)
        self.snippet.setContentsMargins(0, 0, 0, 5)

        self.layout.addWidget(self.snippet)

        # self.bottom_border = QFrame()
        # self.bottom_border.setFrameShape(QFrame.Shape.HLine)
        # self.bottom_border.setFrameShadow(QFrame.Shadow.Sunken)

        # self.layout.addWidget(self.bottom_border)

    def setFromResult(self, r: SearchResult):
        if r['ref'] is not None:
            self.setTitle(f"{r['ref']} {r['title']}")
        else:
            self.setTitle(r['title'])

        if r['author'] is not None:
            self.setAuthor(r['author'])

        self.setSnippet(r['snippet'])

        if r['uid'] is not None:
            self.setDetails(r['uid'])

    def setTitle(self, text: str):
        if len(text.strip()) == 0:
            text = "(Untitled)"
        text = f"<b>{text}</b>"
        self.title.setText(text)

    def setAuthor(self, text: str):
        if len(text.strip()) == 0:
            self.author.clear()
            return
        text = f"<b>{text}</b>"
        self.author.setText(text)

    def setSnippet(self, text):
        self.snippet.setText(text)

    def setDetails(self, text: str):
        if len(text.strip()) == 0:
            self.details.clear()
            return
        text = f"<i>{text}</i>"
        self.details.setText(text)



from functools import partial
from typing import List, Optional
import logging as _logging
import json
from PyQt5 import QtWidgets

from PyQt5.QtCore import Qt, QAbstractListModel, QItemSelectionModel
from PyQt5.QtWidgets import QListView, QListWidget, QMessageBox, QPlainTextEdit

from sqlalchemy.sql import func

from ..app.file_doc import FileDoc
from ..app.db import appdata_models as Am
from ..app.db import userdata_models as Um

from ..app.types import AppData, USutta, UDictWord, UMemo

logger = _logging.getLogger(__name__)


class MemoPlainListModel(QAbstractListModel):
    def __init__(self, *args, memos=None, **kwargs):
        super(MemoPlainListModel, self).__init__(*args, **kwargs)
        self.memos = memos or []

    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole:
            fields = json.loads(self.memos[index.row()].fields_json)
            text = "\n".join(fields.values())
            return text

    def rowCount(self, index):
        if self.memos:
            return len(self.memos)
        else:
            return 0


class HasMemosSidebar:
    _app_data: AppData
    front_input: QtWidgets.QPlainTextEdit
    back_input: QtWidgets.QPlainTextEdit
    front: QPlainTextEdit
    back: QPlainTextEdit
    features: List[str] = []
    memos_list: QListView
    _current_sutta: Optional[USutta]
    _current_word: Optional[UDictWord]
    file_doc: Optional[FileDoc]
    db_doc: Optional[Um.Document]
    clear_memo_button: QtWidgets.QPushButton
    remove_memo_button: QtWidgets.QPushButton
    rightside_tabs: QtWidgets.QTabWidget
    memos_tab_idx: int

    def init_memos_sidebar(self):
        self.features.append('memos_sidebar')

        self.model = MemoPlainListModel()
        self.memos_list.setModel(self.model)
        self.sel_model = self.memos_list.selectionModel()

        palette = self.memos_list.palette()

        style = """
QListView::item { border-bottom: 1px solid %s; }
QListView::item:selected { background-color: %s; color: %s; }
        """ % (palette.midlight().color().name(),
               palette.highlight().color().name(),
               palette.highlightedText().color().name())

        self.memos_list.setStyleSheet(style)

        self.show_memo_clear()

        self.connect_memos_sidebar_signals()

    def show_memo_clear(self):
        self.front_input.clear()
        self.back_input.clear()

    def get_memos_for_sutta(self, sutta: USutta) -> List[UMemo]:
        am_assoc = []
        um_assoc = []

        schema = sutta.metadata.schema

        if schema == 'appdata':

            res = self._app_data.db_session \
                                .query(Am.MemoAssociation) \
                                .filter(
                                    Am.MemoAssociation.associated_table == 'appdata.suttas',
                                    Am.MemoAssociation.associated_id == sutta.id) \
                                .all()
            am_assoc.extend(res)

            res = self._app_data.db_session \
                                .query(Um.MemoAssociation) \
                                .filter(
                                    Um.MemoAssociation.associated_table == 'appdata.suttas',
                                    Um.MemoAssociation.associated_id == sutta.id) \
                                .all()
            um_assoc.extend(res)

        else:

            res = self._app_data.db_session \
                                .query(Um.MemoAssociation) \
                                .filter(
                                    Um.MemoAssociation.associated_table == 'userdata.suttas',
                                    Um.MemoAssociation.associated_id == sutta.id) \
                                .all()
            um_assoc.extend(res)

        memos: List[UMemo] = []

        ids = list(map(lambda x: x.memo_id, am_assoc))

        res = self._app_data.db_session \
                            .query(Am.Memo) \
                            .filter(Am.Memo.id.in_(ids)) \
                            .all()
        memos.extend(res)

        ids = list(map(lambda x: x.memo_id, um_assoc))

        res = self._app_data.db_session \
                            .query(Um.Memo) \
                            .filter(Um.Memo.id.in_(ids)) \
                            .all()
        memos.extend(res)

        return memos

    def get_memos_for_dict_word(self, word: UDictWord) -> List[UMemo]:
        am_assoc = []
        um_assoc = []

        schema = word.metadata.schema

        if schema == 'appdata':

            res = self._app_data.db_session \
                                .query(Am.MemoAssociation) \
                                .filter(
                                    Am.MemoAssociation.associated_table == 'appdata.dict_words',
                                    Am.MemoAssociation.associated_id == word.id) \
                                .all()
            am_assoc.extend(res)

            res = self._app_data.db_session \
                                .query(Um.MemoAssociation) \
                                .filter(
                                    Um.MemoAssociation.associated_table == 'appdata.dict_words',
                                    Um.MemoAssociation.associated_id == word.id) \
                                .all()
            um_assoc.extend(res)

        else:

            res = self._app_data.db_session \
                                .query(Um.MemoAssociation) \
                                .filter(
                                    Um.MemoAssociation.associated_table == 'userdata.dict_words',
                                    Um.MemoAssociation.associated_id == word.id) \
                                .all()
            um_assoc.extend(res)

        memos: List[UMemo] = []

        ids = list(map(lambda x: x.memo_id, am_assoc))

        res = self._app_data.db_session \
                            .query(Am.Memo) \
                            .filter(Am.Memo.id.in_(ids)) \
                            .all()
        memos.extend(res)

        ids = list(map(lambda x: x.memo_id, um_assoc))

        res = self._app_data.db_session \
                            .query(Um.Memo) \
                            .filter(Um.Memo.id.in_(ids)) \
                            .all()
        memos.extend(res)

        return memos

    def update_memos_list_for_sutta(self, sutta: USutta):
        memos = self.get_memos_for_sutta(sutta)
        if memos:
            self.model.memos = memos
        else:
            self.model.memos = []

        self.update_memos_list()

    def update_memos_list_for_dict_word(self, word: UDictWord):
        memos = self.get_memos_for_dict_word(word)
        if memos:
            self.model.memos = memos
        else:
            self.model.memos = []

        self.update_memos_list()

    def update_memos_list_for_document(self, file_doc: FileDoc, db_doc: Um.Document):
        memos = self.get_memos_for_document_page(file_doc, db_doc)
        if memos:
            self.model.memos = memos
        else:
            self.model.memos = []

        self.update_memos_list()

    def update_memos_list(self):
        self.memos_list.clearSelection()
        self.show_memo_clear()

        hits = len(self.model.memos)
        if hits > 0:
            self.rightside_tabs.setTabText(self.memos_tab_idx, f"Memos ({hits})")
        else:
            self.rightside_tabs.setTabText(self.memos_tab_idx, "Memos")

        self.model.layoutChanged.emit()

    def get_memos_for_document_page(self, file_doc: FileDoc, db_doc: Um.Document) -> List[UMemo]:
        am_assoc = []
        um_assoc = []

        doc_schema = db_doc.metadata.schema

        if doc_schema == 'appdata':

            res = self._app_data.db_session \
                                .query(Am.MemoAssociation) \
                                .filter(
                                    Am.MemoAssociation.associated_table == 'appdata.documents',
                                    Am.MemoAssociation.associated_id == db_doc.id,
                                    Am.MemoAssociation.page_number == file_doc.current_page_number()) \
                                .all()
            am_assoc.extend(res)

            res = self._app_data.db_session \
                                .query(Um.MemoAssociation) \
                                .filter(
                                    Um.MemoAssociation.associated_table == 'appdata.documents',
                                    Um.MemoAssociation.associated_id == db_doc.id,
                                    Um.MemoAssociation.page_number == file_doc.current_page_number()) \
                                .all()
            um_assoc.extend(res)

        else:

            res = self._app_data.db_session \
                                .query(Um.MemoAssociation) \
                                .filter(
                                    Um.MemoAssociation.associated_table == 'userdata.documents',
                                    Um.MemoAssociation.associated_id == db_doc.id,
                                    Um.MemoAssociation.page_number == file_doc.current_page_number()) \
                                .all()
            um_assoc.extend(res)

        memos: List[UMemo] = []

        ids = list(map(lambda x: x.memo_id, am_assoc))

        res = self._app_data.db_session \
                            .query(Am.Memo) \
                            .filter(Am.Memo.id.in_(ids)) \
                            .all()
        memos.extend(res)

        ids = list(map(lambda x: x.memo_id, um_assoc))

        res = self._app_data.db_session \
                            .query(Um.Memo) \
                            .filter(Um.Memo.id.in_(ids)) \
                            .all()
        memos.extend(res)

        return memos

    def get_selected_memo(self) -> Optional[UMemo]:
        a = self.memos_list.selectedIndexes()
        if not a:
            return None

        item = a[0]
        return self.model.memos[item.row()]

    def remove_selected_memo(self):
        a = self.memos_list.selectedIndexes()
        if not a:
            return None

        # Remove from model
        item = a[0]
        memo = self.model.memos[item.row()]
        memo_id = memo.id
        schema = memo.metadata.schema

        del self.model.memos[item.row()]
        self.model.layoutChanged.emit()
        self.memos_list.clearSelection()
        self.show_memo_clear()

        # Remove memo from database

        if schema == 'appdata':
            db_item = self._app_data.db_session \
                .query(Am.Memo) \
                .filter(Am.Memo.id == memo_id) \
                .first()

            assoc_item = self._app_data.db_session \
                .query(Am.MemoAssociation) \
                .filter(Am.MemoAssociation.memo_id == memo_id) \
                .first()
        else:
            db_item = self._app_data.db_session \
                .query(Um.Memo) \
                .filter(Um.Memo.id == memo_id) \
                .first()

            assoc_item = self._app_data.db_session \
                .query(Um.MemoAssociation) \
                .filter(Um.MemoAssociation.memo_id == memo_id) \
                .first()

        self._app_data.db_session.delete(db_item)
        self._app_data.db_session.delete(assoc_item)
        self._app_data.db_session.commit()

    def _handle_memo_select(self):
        memo = self.get_selected_memo()
        if memo:
            self._show_memo(memo)

    def _show_memo(self, memo: UMemo):
        fields = json.loads(memo.fields_json) # type: ignore
        self.front_input.setPlainText(fields['Front'])
        self.back_input.setPlainText(fields['Back'])

    def clear_memo(self):
        self.sel_model.clearSelection()
        self.front_input.clear()
        self.back_input.clear()

    def add_memo_for_sutta(self):
        if self._current_sutta is None:
            return

        self.sel_model.clearSelection()

        front = self.front_input.toPlainText()
        back = self.back_input.toPlainText()

        if len(front) == 0 and len(back) == 0:
            logger.info("Empty content, cancel adding.")
            return

        # Insert database record

        logger.info("Adding new memo")

        deck = self._app_data.db_session.query(Um.Deck).first()
        if deck is None:
            logger.error("Can't find the deck")
            return

        memo_fields = {
            'Front': front,
            'Back': back
        }

        memo = Um.Memo(
            deck_id=deck.id,
            fields_json=json.dumps(memo_fields),
            created_at=func.now(),
        )

        try:
            self._app_data.db_session.add(memo)
            self._app_data.db_session.commit()

            memo_assoc = Um.MemoAssociation(
                memo_id=memo.id,
                associated_table='appdata.suttas',
                associated_id=self._current_sutta.id,
            )

            self._app_data.db_session.add(memo_assoc)
            self._app_data.db_session.commit()

            # Add to model
            if self.model.memos:
                self.model.memos.append(memo)
            else:
                self.model.memos = [memo]

        except Exception as e:
            logger.error(e)

        index = self.model.index(len(self.model.memos) - 1)
        self.memos_list.selectionModel().select(index, QItemSelectionModel.SelectionFlag.Select)

        self.update_memos_list()

    def add_memo_for_dict_word(self):
        if self._current_word is None:
            return

        self.sel_model.clearSelection()

        front = self.front_input.toPlainText()
        back = self.back_input.toPlainText()

        if len(front) == 0 and len(back) == 0:
            logger.info("Empty content, cancel adding.")
            return

        # Insert database record

        logger.info("Adding new memo")

        deck = self._app_data.db_session.query(Um.Deck).first()
        if deck is None:
            logger.error("Can't find the deck")
            return

        memo_fields = {
            'Front': front,
            'Back': back
        }

        memo = Um.Memo(
            deck_id=deck.id,
            fields_json=json.dumps(memo_fields),
            created_at=func.now(),
        )

        try:
            self._app_data.db_session.add(memo)
            self._app_data.db_session.commit()

            schema = self._current_word.metadata.schema

            memo_assoc = Um.MemoAssociation(
                memo_id=memo.id,
                associated_table=f"{schema}.dict_words",
                associated_id=self._current_word.id,
            )

            self._app_data.db_session.add(memo_assoc)
            self._app_data.db_session.commit()

            # Add to model
            if self.model.memos:
                self.model.memos.append(memo)
            else:
                self.model.memos = [memo]

        except Exception as e:
            logger.error(e)

        index = self.model.index(len(self.model.memos) - 1)
        self.memos_list.selectionModel().select(index, QItemSelectionModel.SelectionFlag.Select)

        self.update_memos_list()

    def add_memo_for_document(self):
        if self.file_doc is None:
            return

        self.sel_model.clearSelection()

        front = self.front_input.toPlainText()
        back = self.back_input.toPlainText()

        if len(front) == 0 and len(back) == 0:
            logger.info("Empty content, cancel adding.")
            return

        # Insert database record

        logger.info("Adding new memo")

        deck = self._app_data.db_session.query(Um.Deck).first()
        if deck is None:
            logger.error("Can't find the deck")
            return

        memo_fields = {
            'Front': front,
            'Back': back
        }

        memo = Um.Memo(
            deck_id=deck.id,
            fields_json=json.dumps(memo_fields),
            created_at=func.now(),
        )

        try:
            self._app_data.db_session.add(memo)
            self._app_data.db_session.commit()

            if self.db_doc is not None:

                memo_assoc = Um.MemoAssociation(
                    memo_id=memo.id,
                    associated_table='userdata.documents',
                    associated_id=self.db_doc.id,
                    page_number=self.file_doc.current_page_number(),
                )

                self._app_data.db_session.add(memo_assoc)
                self._app_data.db_session.commit()

            # Add to model
            if self.model.memos:
                self.model.memos.append(memo)
            else:
                self.model.memos = [memo]

        except Exception as e:
            logger.error(e)

        index = self.model.index(len(self.model.memos) - 1)
        self.memos_list.selectionModel().select(index, QItemSelectionModel.SelectionFlag.Select)

        self.update_memos_list()

    def update_selected_memo_fields(self):
        memo = self.get_selected_memo()
        if memo is None:
            return

        fields = {
            'Front': self.front_input.toPlainText(),
            'Back': self.back_input.toPlainText()
        }

        memo.fields_json = json.dumps(fields) # type: ignore

        self._app_data.db_session.commit()
        self.model.layoutChanged.emit()

    def remove_memo_dialog(self):
        memo = self.get_selected_memo()
        if not memo:
            return

        reply = QMessageBox.question(self, # type: ignore
                                     'Remove Memo...',
                                     'Remove this item?',
                                     QMessageBox.Yes | QMessageBox.No,
                                     QMessageBox.No)
        if reply == QMessageBox.Yes:
            self.remove_selected_memo()
            self.update_memos_list()

    def connect_memos_sidebar_signals(self):
        self.sel_model.selectionChanged.connect(partial(self._handle_memo_select))

        self.clear_memo_button \
            .clicked.connect(partial(self.clear_memo))

        self.remove_memo_button \
            .clicked.connect(partial(self.remove_memo_dialog))

        self.front_input \
            .textChanged.connect(partial(self.update_selected_memo_fields))

        self.back_input \
            .textChanged.connect(partial(self.update_selected_memo_fields))



"""Stardict related database funcions
"""

import logging as _logging

from typing import Optional, List, TypedDict

from sqlalchemy.sql import func
from sqlalchemy.dialects.sqlite import insert

from simsapa.app.stardict import DictEntry, StarDictPaths, stardict_to_dict_entries, parse_ifo

from . import appdata_models as Am
from . import userdata_models as Um

logger = _logging.getLogger(__name__)

class DbDictEntry(TypedDict):
    word: str
    definition_plain: str
    definition_html: str
    synonyms: str
    uid: str
    dictionary_id: int

def db_entries(x: DictEntry, dictionary_id: int, dictionary_label: str) -> DbDictEntry:
    # TODO should we check for conflicting uids? generate with meaning count?
    uid = f"{x['word']}/{dictionary_label}".lower()
    return DbDictEntry(
        # copy values
        word = x['word'],
        definition_plain = x['definition_plain'],
        definition_html = x['definition_html'],
        synonyms = ", ".join(x['synonyms']),
        # add missing data
        uid = uid,
        dictionary_id = dictionary_id,
    )

def insert_db_words(db_session, schema_name: str, db_words: List[DbDictEntry], batch_size = 1000):
    inserted = 0

    # TODO: The user can't see this message. Dialog doesn't update while the
    # import is blocking the GUI.
    # self.msg.setText("Importing ...")
    print("Importing ...")

    while inserted <= len(db_words):
        b_start = inserted
        b_end = inserted + batch_size
        words_batch = db_words[b_start:b_end]

        try:
            if schema_name == 'userdata':
                stmt = insert(Um.DictWord).values(words_batch)
            else:
                stmt = insert(Am.DictWord).values(words_batch)

            # update the record if uid already exists
            stmt = stmt.on_conflict_do_update(
                index_elements = [Um.DictWord.uid],
                set_ = dict(
                    word = stmt.excluded.word,
                    word_nom_sg = stmt.excluded.word_nom_sg,
                    inflections = stmt.excluded.inflections,
                    phonetic = stmt.excluded.phonetic,
                    transliteration = stmt.excluded.transliteration,
                    # --- Meaning ---
                    meaning_order = stmt.excluded.meaning_order,
                    definition_plain = stmt.excluded.definition_plain,
                    definition_html = stmt.excluded.definition_html,
                    summary = stmt.excluded.summary,
                    # --- Associated words ---
                    synonyms = stmt.excluded.synonyms,
                    antonyms = stmt.excluded.antonyms,
                    homonyms = stmt.excluded.homonyms,
                    also_written_as = stmt.excluded.also_written_as,
                    see_also = stmt.excluded.see_also,
                )
            )

            db_session.execute(stmt)
            db_session.commit()
        except Exception as e:
            print(e)
            logger.error(e)

        inserted += batch_size
        # self.msg.setText(f"Imported {inserted} ...")
        print(f"Imported {inserted} ...")

def import_stardict_into_db_update_existing(db_session,
                                            schema_name: str,
                                            paths: StarDictPaths,
                                            dictionary_id: int,
                                            label: str,
                                            batch_size = 1000):
    words: List[DictEntry] = stardict_to_dict_entries(paths)
    db_words: List[DbDictEntry] = list(map(lambda x: db_entries(x, dictionary_id, label), words))
    insert_db_words(db_session, schema_name, db_words, batch_size)

def import_stardict_into_db_as_new(db_session,
                                   schema_name: str,
                                   paths: StarDictPaths,
                                   label: Optional[str] = None,
                                   batch_size = 1000):
    # upsert recommended by docs instead of bulk_insert_mappings
    # Using PostgreSQL ON CONFLICT with RETURNING to return upserted ORM objects
    # https://docs.sqlalchemy.org/en/14/orm/persistence_techniques.html#using-postgresql-on-conflict-with-returning-to-return-upserted-orm-objects

    words: List[DictEntry] = stardict_to_dict_entries(paths)
    ifo = parse_ifo(paths)
    title = ifo['bookname']
    if label is None:
        label = title

    # create a dictionary, commit to get its ID
    if schema_name == 'userdata':
        dictionary = Um.Dictionary(
            title = title,
            label = label,
            created_at = func.now(),
        )
    else:
        dictionary = Am.Dictionary(
            title = title,
            label = label,
            created_at = func.now(),
        )

    try:
        db_session.add(dictionary)
        db_session.commit()
    except Exception as e:
        logger.error(e)

    d_id: int = dictionary.id # type: ignore
    db_words: List[DbDictEntry] = list(map(lambda x: db_entries(x, d_id, label), words))

    insert_db_words(db_session, schema_name, db_words, batch_size)



from functools import partial
from typing import Optional
import logging as _logging
from pathlib import Path
import queue
import json

from PyQt5.QtCore import Qt, QPoint, QRect, QUrl, QTimer
from PyQt5.QtGui import QImage, QPixmap, QCloseEvent
from PyQt5.QtWidgets import (QLabel, QMainWindow, QFileDialog, QInputDialog, QAction)

from simsapa import ASSETS_DIR, APP_QUEUES

from ..app.file_doc import FileDoc, PageImage
from ..app.db import appdata_models as Am
from ..app.db import userdata_models as Um

from ..app.types import AppData, USutta
from ..assets.ui.document_reader_window_ui import Ui_DocumentReaderWindow
from .memos_sidebar import HasMemosSidebar
from .links_sidebar import HasLinksSidebar

logger = _logging.getLogger(__name__)


class DocumentReaderWindow(QMainWindow, Ui_DocumentReaderWindow, HasLinksSidebar, HasMemosSidebar):
    def __init__(self, app_data: AppData, parent=None) -> None:
        super().__init__(parent)
        self.setupUi(self)

        self.features = []
        self._app_data: AppData = app_data

        self.queue_id = 'window_' + str(len(APP_QUEUES))
        APP_QUEUES[self.queue_id] = queue.Queue()
        self.messages_url = f'{self._app_data.api_url}/queues/{self.queue_id}'

        self.graph_path: Path = ASSETS_DIR.joinpath(f"{self.queue_id}.html")

        self.timer = QTimer()
        self.timer.timeout.connect(self.handle_messages)
        self.timer.start(300)

        self._ui_setup()
        self._connect_signals()

        self.init_memos_sidebar()
        self.init_links_sidebar()

        self.statusbar.showMessage("Ready", 3000)

    def closeEvent(self, event: QCloseEvent):
        if self.queue_id in APP_QUEUES.keys():
            del APP_QUEUES[self.queue_id]

        if self.graph_path.exists():
            self.graph_path.unlink()

        event.accept()

    def handle_messages(self):
        if self.queue_id in APP_QUEUES.keys():
            try:
                s = APP_QUEUES[self.queue_id].get_nowait()
                data = json.loads(s)
                if data['action'] == 'show_sutta':
                    self._show_sutta_from_message(data['arg'])

                APP_QUEUES[self.queue_id].task_done()
            except queue.Empty:
                pass

    def _ui_setup(self):
        self.status_msg = QLabel("")
        self.statusbar.addPermanentWidget(self.status_msg)

        self.links_tab_idx = 0
        self.memos_tab_idx = 1

        self.file_doc: Optional[FileDoc] = None
        self.db_doc: Optional[Um.Document] = None
        self.content_page_image: Optional[PageImage] = None

        self.selecting = False
        self.selected_text = None
        self.select_start_point: Optional[QPoint] = None
        self.select_end_point: Optional[QPoint] = None
        self.select_rectangle: Optional[QRect] = None

    def open_file_dialog(self):
        file_path, _ = QFileDialog.getOpenFileName(
            None,
            "Open File...",
            "",
            "PDF or Epub Files (*.pdf *.epub)")

        if len(file_path) != 0:
            self.open_doc(file_path)

    def open_doc(self, path):
        self.db_doc = self._app_data.db_session \
                                    .query(Um.Document) \
                                    .filter(Um.Document.filepath == path) \
                                    .first()

        if self.db_doc is not None and self.file_doc is not None:
            self.update_memos_list_for_document(self.file_doc, self.db_doc)

        self.file_doc = FileDoc(path)
        self.doc_go_to_page(1)

    def doc_show_current(self):
        if self.file_doc is not None:
            self.doc_go_to_page(self.file_doc.current_page_number())

    def doc_go_to_page(self, page: int):
        if not self.file_doc:
            return

        # logger.info(f"doc_go_to_page({page})")

        page_count = self.file_doc.number_of_pages()

        if self.file_doc is None or page < 1 or page > page_count:
            return

        self.page_current_of_total.setText(f"{page} of {page_count}")

        if self.file_doc:
            self.file_doc.go_to_page(page)

            page_img = self.file_doc.current_page_image
            self.content_page_image = page_img

            if page_img:
                img = QImage(page_img.image_bytes,
                             page_img.width,
                             page_img.height,
                             page_img.stride,
                             QImage.Format.Format_RGB888)
                self.content_page.setPixmap(QPixmap.fromImage(img))

                self.content_page.mousePressEvent = self._select_start
                self.content_page.mouseMoveEvent = self._select_move
                self.content_page.mouseReleaseEvent = self._select_release

        if self.db_doc:
            self.update_memos_list_for_document(self.file_doc, self.db_doc)
            self.show_network_graph()

    def show_network_graph(self):
        if self.file_doc is None or self.db_doc is None:
            return

        self.generate_graph_for_document(self.file_doc, self.db_doc, self.queue_id, self.graph_path, self.messages_url)
        self.content_graph.load(QUrl('file://' + str(self.graph_path.absolute())))

    def _show_sutta_from_message(self, info):
        sutta: Optional[USutta] = None

        if info['table'] == 'appdata.suttas':
            sutta = self._app_data.db_session \
                .query(Am.Sutta) \
                .filter(Am.Sutta.id == info['id']) \
                .first()

        elif info['table'] == 'userdata.suttas':
            sutta = self._app_data.db_session \
                .query(Um.Sutta) \
                .filter(Um.Sutta.id == info['id']) \
                .first()

        self._app_data.sutta_to_open = sutta
        self.action_Sutta_Search.activate(QAction.ActionEvent.Trigger)

    def _select_start(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            self.selecting = True
            self.select_start_point = QPoint(event.pos())

    def _select_move(self, event):
        if not self.selecting:
            return
        if self.select_start_point is None:
            return
        self.select_end_point = QPoint(event.pos())
        self.select_rectangle = QRect(self.select_start_point, self.select_end_point).normalized()
        if self.file_doc and self.select_rectangle:
            self.selected_text = self.file_doc.get_selection_text(self.select_rectangle)
            self.file_doc.paint_selection_on_current(self.select_rectangle)
            self.doc_show_current()

    def _select_release(self, event):
        self._select_move(event)
        self.selecting = False

    def _previous_page(self):
        if not self.file_doc:
            return
        page_nr = self.file_doc.current_page_number() - 1
        if page_nr > 0:
            self.file_doc.go_to_page(page_nr)
            self._upd_current_page_input(page_nr)

    def _next_page(self):
        if not self.file_doc:
            return
        page_nr = self.file_doc.current_page_number() + 1
        if page_nr <= self.file_doc.number_of_pages():
            self.file_doc.go_to_page(page_nr)
            self._upd_current_page_input(page_nr)

    def _beginning(self):
        if self.file_doc and self.file_doc.current_page_number() != 1:
            self.file_doc.go_to_page(1)
            self._upd_current_page_input(1)

    def _end(self):
        if self.file_doc and self.file_doc.current_page_number() != self.file_doc.number_of_pages():
            self.file_doc.go_to_page(self.file_doc.number_of_pages())
            self._upd_current_page_input(self.file_doc.number_of_pages())

    def _upd_current_page_input(self, n):
        self.current_page_input.setValue(n)
        self.current_page_input.clearFocus()

    def _go_to_page_dialog(self):
        if not self.file_doc:
            return

        n, ok = QInputDialog.getInt(self, "Go to Page...", "Page:", 1, 1, self.file_doc.number_of_pages(), 1)
        if ok:
            self._upd_current_page_input(n)

    def _go_to_page_input(self):
        n = self.current_page_input.value()
        self.doc_go_to_page(n)

    def _zoom_out(self):
        if self.file_doc is None:
            return
        if self.file_doc._zoom is None:
            return
        self.file_doc.set_zoom(self.file_doc._zoom - 0.1)
        self.doc_show_current()

    def _zoom_in(self):
        if self.file_doc is None:
            return
        if self.file_doc._zoom is None:
            return
        self.file_doc.set_zoom(self.file_doc._zoom + 0.1)
        self.doc_show_current()

    def _connect_signals(self):
        self.action_Close_Window \
            .triggered.connect(partial(self.close))

        self.action_Previous_Page \
            .triggered.connect(partial(self._previous_page))
        self.action_Next_Page \
            .triggered.connect(partial(self._next_page))

        self.action_Beginning \
            .triggered.connect(partial(self._beginning))
        self.action_End \
            .triggered.connect(partial(self._end))

        self.action_Go_to_Page \
            .triggered.connect(partial(self._go_to_page_dialog))

        self.action_Zoom_Out \
            .triggered.connect(partial(self._zoom_out))
        self.action_Zoom_In \
            .triggered.connect(partial(self._zoom_in))

        self.current_page_input \
            .valueChanged.connect(partial(self._go_to_page_input))

        self.add_memo_button \
            .clicked.connect(partial(self.add_memo_for_document))
